# City Attack
# (c) Paul Furber 2002 <paulf@gam.co.za>
# released under the GNU General Public License

# After reading a very thought-provoking piece on component abstraction
# by Geoff Howland on www.ludumdare.com, I thought I would try his
# experiment - concentrate on logic first to get a working game.
# This is the result: an exceptionally simple game that was popular
# with a number of home computers in the 80s. Press the space bar to drop
# a bomb. If you land the city is rebuilt. If you crash, you get to try again.
# Esc to quit.
# Most handy bits of code are borrowed from Shandy Brown's monkeystomp and the
# original chimp punch example that comes with pygame.

# Conclusion: Geoff Howland is absolutely right in his article - a must read
# for developers at http://ludumdare.com/articles/?link=v&arid=6. I've only
# done his Step 1 but I can see that further stages will follow in similar
# fashion. This game could be jazzed up very easily with sound, better gfx,
# a nicer background, more features and taken a lot further than it is. And
# if you think using sprite collision is overkill, you would be right - I was
# just teaching myself how to use the sprite class. It's also a nice code
# base to reuse: game object with some players, image loading, sprite collide
# and simple movement etc. For excellent key handling code, check out Shandy's
# monkeystomp with uses more then just a single key :)
# Dev time: around 4 hours with some interruptions

# Hope this helps anyone wanting to take something from Geoff's article and
# apply it to the joy that is pygame development.

import pygame, os
import random
from pygame.locals import *

if not pygame.mixer: print 'Warning, sound disabled'

#-----------------------------------------------------------------------------
def allKUPs(x): return x.type == KEYUP
#-----------------------------------------------------------------------------
def load_png(name):
    fullname = os.path.join('data', name)
    try:
        image = pygame.image.load(fullname)
	if image.get_alpha is None:
		image = image.convert()
	else:
		image = image.convert_alpha()
    except pygame.error, message:
        print 'Cannot load image:', fullname
        raise SystemExit, message
    return image, image.get_rect()
#-----------------------------------------------------------------------------
def load_sound(name):
    class NoneSound:
        def play(self): pass
    if not pygame.mixer or not pygame.mixer.get_init():
        return NoneSound()
    fullname = os.path.join('data', name)
    try:
        sound = pygame.mixer.Sound(fullname)
    except pygame.error, message:
        print 'Cannot load sound:', fullname
        raise SystemExit, message
    return sound
#-----------------------------------------------------------------------------
class SimpleSprite(pygame.sprite.Sprite):
	def __init__(self, name=None):
		pygame.sprite.Sprite.__init__(self) #call Sprite initializer
 		if name:
			self.image, self.rect = load_png(name)
		else:
			self.image = pygame.Surface( (100,100) )
			self.image.fill( (200,0,0) )
			self.rect = self.image.get_rect()
	def update(self):
		pass
	def set_image(self, newSurface, newRect=None):
		self.image = newSurface
		if newRect:
			self.rect  = newRect
		else:
			self.rect  = newSurface.get_rect()

#-----------------------------------------------------------------------------
class Plane(pygame.sprite.Sprite):
    def __init__(self, game):
        pygame.sprite.Sprite.__init__(self) #call Sprite initializer
        self.image, self.rect = load_png('plane.png')
	self.rightFacingImg = self.image
	self.leftFacingImg = pygame.transform.flip( self.image, 1, 0)

        self.direction = 0    #left to right
        self.increment = 32
        self.oldPos = self.rect
        self.game = game
        self.listenKeys = {}

    def update(self):
        self.check_land()
        self.move()
        if self.direction == 0:
            self.image = self.rightFacingImg
        else:
            self.image = self.leftFacingImg
        self.check_crash()
        

    def check_crash(self):
        global citysprites
        crash_list = pygame.sprite.spritecollide(self, citysprites, 0)
        if len(crash_list) is not 0:
            self.game.crash(self)

    def move(self):
        self.oldPos = self.rect
        self.rect = self.rect.move(self.increment, 0)
        if self.rect.right > 640:
            self.rect.top += 16
            self.increment = -32
            self.direction = 1
        if self.rect.left < 0:
            self.rect.top += 16
            self.increment = 32
            self.direction = 0

    def signal_key( self, event, remainingEvents ):
		if self.listenKeys.has_key( event.key ) \
		  and event.type is KEYDOWN:
			self.listenKeys[event.key]( remainingEvents )

    def check_land(self):
        if (self.rect.top == 464) and (self.rect.left == 32):
            self.game.land()

#-----------------------------------------------------------------------------
class Bomb(pygame.sprite.Sprite):
    def __init__(self, rect, game):
        pygame.sprite.Sprite.__init__(self) #call Sprite initializer
        self.increment = 16
        self.image, self.rect = load_png("bomb.png")
        if rect is not None:
            self.rect = rect
        self.game = game

    def update(self):
        self.move()

    def move(self):
        self.rect = self.rect.move(0, self.increment)
        if self.rect.top==480:
            self.game.miss()
        else:
            self.check_hit()

    def check_hit(self):
        global citysprites
        collide_list = pygame.sprite.spritecollide(self, citysprites,0)
        if len(collide_list) is not 0:
            self.game.hit()
        
#-----------------------------------------------------------------------------
class City:
    def __init__(self, difficulty):
        global citysprites
        global block
        for i in range(18):
            for j in range(random.randrange(0,difficulty*5)):
                block = SimpleSprite("block.png")
                block.rect.move_ip(((i+1)*32),480-(j*16))
                citysprites.add(block)

    def clear(self):
        global citysprites
        citysprites.empty()

#-----------------------------------------------------------------------------

class cityGame:
    def __init__(self, charGroup):
        self.charGroup = charGroup

        self.difficulty = 2
        self.listenKeys = {K_SPACE: self.drop_bomb}
        self.city = City(self.difficulty)

        self.plane = Plane(self)
        self.charGroup.add( [self.plane] )
        self.bomb = None
        

    def crash(self, oldPlane):
        self.plane.kill()
        self.plane = Plane(self)
        self.charGroup.add ([self.plane])
        
    def land(self):
        self.difficulty += 1
        if self.difficulty > 5:
            self.difficulty = 5
        self.plane.kill()
        self.plane = Plane(self)
        self.charGroup.add([self.plane])
        self.city.clear()        
        self.city = City(self.difficulty)

    def drop_bomb(self):
        if self.bomb is None:
            self.bomb = Bomb(self.plane.rect, self)
            self.charGroup.add([self.bomb])

    def miss(self):
        self.bomb.kill()
        self.bomb = None

    def hit(self):
        global citysprites
        dam_group = pygame.sprite.Group()
        colliderects = []
        for i in range(3):
            for j in range((self.bomb.rect.bottom+16-self.plane.rect.top)/16):
                rect = Rect((self.bomb.rect.left-32+i*32,\
                                          self.bomb.rect.bottom-j*16),\
                                         (32,16))
                colliderects.append(rect)

        for rect in colliderects:
            sprite = SimpleSprite()
            sprite.rect = rect
            dam_group.add(sprite)

        list = pygame.sprite.groupcollide(dam_group, citysprites, 1,1)
        
        self.bomb.kill()
        self.bomb = None
        
	
    def signal_key( self, event, remainingEvents ):
        if self.listenKeys.has_key( event.key ):
            self.listenKeys[event.key]()
        else:
            self.plane.signal_key( event, remainingEvents )
		

#-----------------------------------------------------------------------------
def run():
    global screen
    pygame.init()
    screen = pygame.display.set_mode((640, 480))
    pygame.display.set_caption('City Attack 1.0')

    global background
    background, tmp_rect = load_png('background.png')
    
    screen.blit(background, (0, 0))
    pygame.display.flip()

    allsprites = pygame.sprite.RenderUpdates()
    global citysprites
    citysprites= pygame.sprite.RenderUpdates()

    global block
    block = None
    
    game = cityGame(allsprites)
    
    clock = pygame.time.Clock()
    random.seed(1234579)

    oldfps = 0
    while 1:
        clock.tick(10)
        newfps = int(clock.get_fps())
        if newfps is not oldfps:
            oldfps = newfps

        oldEvents = []
        remainingEvents = pygame.event.get()
        for event in remainingEvents:
            oldEvents.append( remainingEvents.pop(0) )
            upKeys = filter( allKUPs, remainingEvents )

            if event.type == QUIT:
                return
            elif event.type == KEYDOWN and event.key == K_ESCAPE:
                return
            elif event.type == KEYDOWN or event.type == KEYUP:
                game.signal_key( event, upKeys )

	#Draw Everything
        allsprites.clear( screen, background)
        citysprites.clear(screen, background)

        allsprites.update()

        changedRects2 = allsprites.draw(screen)
        changedRects3 = citysprites.draw(screen)
        pygame.display.update(changedRects2+changedRects3)



if __name__ == "__main__":
    run()
